/* ++++++++++
	driver.c
	A skeletal device driver
+++++ */

#include <drivers/KernelExport.h>
#include <drivers/Drivers.h>
#include <support/Errors.h>
#include <string.h>

#define	DRIVER_NAME	"mem"
#define	DEVICE_NAME	"misc/mem"

int32	api_version = B_CUR_DRIVER_API_VERSION;

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */

_EXPORT status_t
init_hardware (void)
{
	dprintf (DRIVER_NAME ": init_hardware()\n");
	return B_OK;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */

_EXPORT status_t
init_driver (void)
{
	dprintf (DRIVER_NAME ": init_driver()\n");
	return B_OK;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */

_EXPORT void
uninit_driver (void)
{
	dprintf (DRIVER_NAME ": uninit_driver()\n");
}

	
/* ----------
	my_device_open - handle open() calls
----- */

static status_t
my_device_open (const char *name, uint32 flags, void** cookie)
{
	dprintf (DRIVER_NAME ": open(\"%s\")\n", name);
	return B_OK;
}


/* ----------
	my_device_read - handle read() calls
----- */

static status_t
my_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	area_id	area;
	void	*paddr, *vaddr;
	size_t	size, bias;

	dprintf (DRIVER_NAME ": read():\n");
	/* intended to work with 64-bit pointer too, but nobody knows... */
	paddr = (void *) (position & ~((off_t) B_PAGE_SIZE - 1));
	bias = position - (off_t) paddr;
	size = (bias + *num_bytes + B_PAGE_SIZE - 1) & ~((size_t) B_PAGE_SIZE - 1);
	area = map_physical_memory ("hogehoge", paddr, size, 
		B_ANY_KERNEL_ADDRESS, B_READ_AREA, &vaddr);
	if (area < 0)
	{
		*num_bytes = 0;
		return B_IO_ERROR;
	}
	memcpy (buf, (caddr_t) vaddr + bias, *num_bytes);
	delete_area (area);
	return B_OK;
}


/* ----------
	my_device_write - handle write() calls
----- */

static status_t
my_device_write (void* cookie, off_t position, const void* buf, size_t* num_bytes)
{
	area_id	area;
	void	*paddr, *vaddr;
	size_t	size, bias;

	dprintf (DRIVER_NAME ": read():\n");
	paddr = (void *) (position & ~((off_t) B_PAGE_SIZE - 1));
	bias = position - (off_t) paddr;
	size = (bias + *num_bytes + B_PAGE_SIZE - 1) & ~((size_t) B_PAGE_SIZE - 1);
	area = map_physical_memory ("hogehoge", paddr, size, 
		B_ANY_KERNEL_ADDRESS, B_WRITE_AREA, &vaddr);
	if (area < 0)
	{
		*num_bytes = 0;
		return B_IO_ERROR;
	}
	memcpy ((caddr_t) vaddr + bias, buf, *num_bytes);
	delete_area (area);
	return B_OK;
}


/* ----------
	my_device_control - handle ioctl calls
----- */

static status_t
my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	dprintf (DRIVER_NAME ": ioctl(%d)\n", op);
	return B_ERROR;
}


/* ----------
	my_device_close - handle close() calls
----- */

static status_t
my_device_close (void* cookie)
{
	dprintf (DRIVER_NAME ": close()\n");
	return B_OK;
}


/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */

static status_t
my_device_free (void* cookie)
{
	dprintf (DRIVER_NAME ": free()\n");
	return B_OK;
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	DEVICE_NAME,
	NULL
};

/* -----
	function pointers for the device hooks entry points
----- */

static device_hooks my_device_hooks = {
	my_device_open, 			/* -> open entry point */
	my_device_close, 			/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

_EXPORT const char**
publish_devices()
{
	dprintf (DRIVER_NAME ": publish_devices()\n");
	return my_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

static int
lookup_device_name (const char *name)
{
	int	i;

	for (i = 0; my_device_name [i] != NULL; i++)
		if (strcmp (name, my_device_name [i]) == 0)
			return i;
	return -1;
}

_EXPORT device_hooks*
find_device(const char* name)
{
	dprintf (DRIVER_NAME ": find_device(\"%s\")\n", name);
	if (lookup_device_name (name) < 0)
		return NULL;
	return &my_device_hooks;
}

